//#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "sdl.h"

// screen attributes
#define SCREEN_WIDTH	640
#define SCREEN_HEIGHT	480
#define SCREEN_BPP		16 // in windowed mode this will be the desktop bpp

// fullscreen
//#define SCREEN_FLAGS	SDL_ANYFORMAT|SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_FULLSCREEN
// windowed
#define SCREEN_FLAGS	SDL_ANYFORMAT|SDL_HWSURFACE|SDL_DOUBLEBUF

// size of our tiles
#define TILE_WIDTH		48
#define TILE_HEIGHT		48

// size of our sprite
#define SPRITE_WIDTH	32
#define SPRITE_HEIGHT	32

// size of the map -arbitrary 
#define MAP_WIDTH		200
#define MAP_HEIGHT		150

// make it easier to understand what direction we are going in
#define GAME_UP			0
#define GAME_RIGHT		1
#define GAME_DOWN		2
#define GAME_LEFT		3

// each direction has 2 frames - old school!
#define FRAME_STAND		0
#define FRAME_WALK		1

// change the animation speed
// this is in milliseconds - play with it a little to see the difference
#define ANIMATION_SPEED	225

// a simple camera structure
typedef struct camera
{
	float	x; // the x pixel position of the camera
	float	y; // the y pixel position of the camera
}CAMERA;

// a very VERY simple tile structure
// lots of room for improvement
typedef struct tile
{
	// this is the source rectangle of the tile image.
	// since we only use one surface we don't keep track of it here
	// but if we had more than one tile surface to deal with then it might be
	// a good idea to keep a reference to the surface it belongs to here
	SDL_Rect	rect;

	// is the tile walkable or not?
	bool		walk;
}TILE;

// a simple sprite structure
typedef struct sprite
{
	// this is 2d array of source rectangles representing each direction and
	// the number of animation frames per direction. again we only use one
	// surface in this example and keeping a reference to the surface would be
	// a good idea if more than one sprite surface was in use.
	SDL_Rect	rect[4][2];

	// using floats for position instead of integers reduces the "jitter" effect
	// from time based movement where rounding makes things look choppy
	float		pos_x;
	float		pos_y;

	// an anchor point is used to represent the pixel on the sprite that is
	// its actual location. in most cases this is a point near the sprites
	// feet. so when we draw a sprite we can draw it regardless of it's size
	// in the same spot.
	int			anchor_x;
	int			anchor_y;

	// the collision radius is for circular collision detection which
	// is not yet implemented here.
	int			collision_radius;


	int			dir;		// what direction is the sprite facing?
	int			frame;		// what frame of animation are we using?
	bool		moving;		// is the sprite moving?
	bool		was_moving;	// was it moving the last time we updated it?
	int			timer;		// time left before we change frames (milliseconds)
}SPRITE;

// global variables. not exactly the best way to do this but for the purpose
// of this short example it is fine.
SDL_Surface		*g_screen;	// the screen surface
SDL_Surface		*g_image;	// the tile surface - these could use better names
SDL_Surface		*g_image2;	// the sprite surface
TILE			g_tiles[4];	// the tiles (only 4 on the bitmap)
int				*g_tilemap;	// our tile map
CAMERA			g_camera;	// the camera
bool			g_keys[4];	// keep track of the keys we want (up,right,down,left)
SPRITE			g_sprite;	// the sprite

// function aptly named to convert a set of world coordinates to screen coordinates
void world_to_screen( float *worldx, float *worldy )
{
	float screenx, screeny;

	screenx = *worldx - g_camera.x;
	screeny = *worldy - g_camera.y;

	if ( screenx < 0 )
		screenx += MAP_WIDTH*TILE_WIDTH;

	if ( screeny < 0 )
		screeny += MAP_HEIGHT*TILE_HEIGHT;

	*worldx = screenx;
	*worldy = screeny;
}

// function aptly named to convert a set of screen coordinates to world coordinates
void screen_to_world( float *screenx, float *screeny )
{
	float worldx, worldy;

	// the camera is basicly an offset to tell us what pixel of our map is the upper left
	// corner of the screen. by adding the camera position to the screen position we
	// get the world position
	worldx = *screenx + g_camera.x;
	worldy = *screeny + g_camera.y;

	// this next little bit gives us that old school NES wrap around effect
	// comment it out for borders but then you need to add limits so you never
	// go out of bounds on the map array
	if ( worldx > MAP_WIDTH*TILE_WIDTH )
		worldx -= MAP_WIDTH*TILE_WIDTH;

	if ( worldy > MAP_HEIGHT*TILE_HEIGHT )
		worldy -= MAP_HEIGHT*TILE_HEIGHT;

	*screenx = worldx;
	*screeny = worldy;
}

// function to load the tile surface and set up the tile objects
void load_tiles( const char *bitmap )
{
	SDL_Surface *temp;
	int x, y, tile;

	// load the tile images
	// TODO: error checking - I didn't bother but it is always good practice
	temp = SDL_LoadBMP( "tiles.bmp" );
	// convert it to the display format
	g_image = SDL_DisplayFormat( temp );
	// free the temporary surface
	SDL_FreeSurface( temp );
	temp = NULL;

	tile = 0;
	// this is not the recommended way to do this since it relies on
	// hard coded numbers such as the number of tiles in the g_tiles array
	// however fleshing this out a bit isn't very difficult once you understand
	// the concept.
	for ( y = 0; y < g_image->h; y += TILE_HEIGHT )
	{
		for ( x = 0; x < g_image->w; x += TILE_WIDTH )
		{
			g_tiles[tile].walk = TRUE;

			g_tiles[tile].rect.x = x;
			g_tiles[tile].rect.y = y;
			g_tiles[tile].rect.w = TILE_WIDTH;
			g_tiles[tile].rect.h = TILE_HEIGHT;
			tile++;
		}
	}

	// make the water tile no walk
	g_tiles[2].walk = FALSE;
}

// here is where you would load your tilemap from a file
// I chose to randomly create one instead since file I/O
// can be a very long and drawn out thing.
void load_tilemap( int width, int height )
{
	// create the map array
	// I chose a 1 dimensional array since C does this internally
	// anyway and quite frankly it is easier to do dynamicly
	g_tilemap = new int[width*height];

	// create a random tilemap
	for ( int y = 0; y < height; y++ )
	{
		for ( int x = 0; x < width; x++ )
		{
			int tile;

			// make the chance of a water tile a little less
			// by making it have to randomly choose it twice in a row
			// this method works for this example but wouldn't be very
			// effective in a real situation
			if ( (tile = rand() % 4) == 2 )
				tile = rand() % 4;

			// set the tilemap index to the correct tile index
			g_tilemap[x + (y * width)] = tile;
		}
	}
}

// get a pointer to the tile at a specific location
// this particular function will never go out of bounds
TILE *get_tile( int x, int y )
{
	// again, we use the wrap thing here to make sure we never go out of bounds
	// there are more elegant solutions but for this type of game this is fine
	if ( x >= MAP_WIDTH )
		x = x % MAP_WIDTH;

	if ( y >= MAP_HEIGHT )
		y = y % MAP_HEIGHT;

	// and the winner is...
	return &g_tiles[ g_tilemap[x + ( y * MAP_WIDTH )] ];
}

// load the sprite image and set up the attributes
void load_sprite( char *bitmap )
{
	SDL_Surface *temp;
	int dir, frame;

	// Load the bitmap to a surface 
	temp = SDL_LoadBMP( bitmap );
	// convert it to the display format
	g_image2 = SDL_DisplayFormat( temp );
	// free up the temp surface
	SDL_FreeSurface( temp );
	temp = NULL;

	// set the transparent color of the sprite
	// 255,0,255 is very common for this since it is almost never used
	SDL_SetColorKey( g_image2, SDL_SRCCOLORKEY, SDL_MapRGB( g_image2->format, 255, 0, 255 ) );

	// since the sprite is layed out in the bitmap a certain way this loop
	// makes it easy to set the sprite source rectangles and attributes
	for ( dir = 0; dir < 4; dir++ )
	{
		for ( frame = 0; frame < 2; frame++ )
		{
			// set the source rect - pretty self explanitory
			g_sprite.rect[dir][frame].x = frame*SPRITE_WIDTH;
			g_sprite.rect[dir][frame].y = dir*SPRITE_HEIGHT;
			g_sprite.rect[dir][frame].w = SPRITE_WIDTH;
			g_sprite.rect[dir][frame].h = SPRITE_HEIGHT;

			// I chose for the anchor point to be
			// centered horizontaly and 4 pixels up from the lowest
			// point in the source rectangle. It just so happens that
			// this is the general placement of the feet of the sprite
			// I am using :)
			g_sprite.anchor_x = SPRITE_WIDTH/2;
			g_sprite.anchor_y = SPRITE_HEIGHT - 4;

			// the collision radius is 8 pixels in any direction
			// from the anchor point.
			g_sprite.collision_radius = 8;

			// we start in the middle of the screen
			g_sprite.pos_x = SCREEN_WIDTH/2;
			g_sprite.pos_y = SCREEN_HEIGHT/2;

			// we put our best foot forward and default to standing facing down
			g_sprite.dir = GAME_DOWN;
			g_sprite.frame = FRAME_STAND;

			// we were not and are not moving - yet
			g_sprite.moving = FALSE;
			g_sprite.was_moving = FALSE;

			// reset the animation timer to the animation speed
			g_sprite.timer = ANIMATION_SPEED;
		}
	}
}

// check if a point is on a walkable tile or not
bool check_tile_collision( int x, int y )
{
	TILE *tile = get_tile( x/TILE_WIDTH, y/TILE_HEIGHT );

	if ( !tile || !tile->walk )
		return FALSE;

	return TRUE;
}

// move a sprite based on delta amount of time passing
void move_sprite( float delta )
{
	float x,y;
	float speed = 75.0f; // arbitrary - play with it to find what you want
	float off_x = 0.0f, off_y = 0.0f;

	delta /= 1000; 

	// start with where we are at
	x = g_sprite.pos_x;
	y = g_sprite.pos_y;

	// off_x and off_y are how much we will move in any one direction
	// see which keys are down and update the off_x and off_y accordingly
	// this is a simple way to read input from the user and isn't meant for
	// a real applictation, only meant as a proof of concept.
	if ( g_keys[GAME_UP] )
	{
		off_y -= speed*delta;
		g_sprite.dir = GAME_UP;
		g_sprite.moving = TRUE;
	}
	else
	if ( g_keys[GAME_DOWN] )
	{
		off_y += speed*delta;
		g_sprite.dir = GAME_DOWN;
		g_sprite.moving = TRUE;
	}
	else
	if ( g_keys[GAME_LEFT] )
	{
		off_x -= speed*delta;
		g_sprite.dir = GAME_LEFT;
		g_sprite.moving = TRUE;
	}
	else
	if ( g_keys[GAME_RIGHT] )
	{
		off_x += speed*delta;
		g_sprite.dir = GAME_RIGHT;
		g_sprite.moving = TRUE;
	}
	else // none are down so stop
	{
		g_sprite.moving = FALSE;
		g_sprite.frame = 0;
	}

	// manipulate our saved x position
	x += off_x;

	// make sure it is on the map
	if ( x < 0 )
		x += (MAP_WIDTH*TILE_WIDTH);
	else
	if ( x >= (MAP_WIDTH*TILE_WIDTH) )
		x -= (MAP_WIDTH*TILE_WIDTH);


	// manipulate our y position
	y += off_y;

	// make sure it is on the map
	if ( y < 0 )
		y += (MAP_HEIGHT*TILE_HEIGHT);
	else
	if ( y >= (MAP_HEIGHT*TILE_HEIGHT) )
		y -= (MAP_HEIGHT*TILE_HEIGHT);

	// check if the new position is legal
	// if not, then return without moving
	if ( !check_tile_collision( (int)x, (int)y ) )
		return;

	// yay! we moved!
	g_sprite.pos_x = x;
	g_sprite.pos_y = y;
}

// update a sprite's state in the world
void update_sprite( float delta )
{
	// move the sprite
	move_sprite( delta );

	// if the sprite is moving for the first time
	// make it take a first step instead of sliding
	if ( g_sprite.moving && !g_sprite.was_moving )
	{
		g_sprite.was_moving = TRUE;
		g_sprite.timer = ANIMATION_SPEED;
		g_sprite.frame = FRAME_WALK;
	}
	else // if the sprite has stopped change the frame to standing
	if ( !g_sprite.moving && g_sprite.was_moving )
	{
		g_sprite.was_moving = FALSE;
		g_sprite.frame = FRAME_STAND;
	}

	// if the sprite is moving now, update the frame according to the animation speed.
	if ( g_sprite.moving )
	{
		g_sprite.timer -= (int)delta;

		if ( g_sprite.timer <= 0 )
		{
			++g_sprite.frame %= 2;
			g_sprite.timer += ANIMATION_SPEED;
		}
	}
}

// umm.... I think the name says it all
void draw_sprite()
{
	float x, y;
	SDL_Rect dest;

	// notice how the anchor point is used to draw the sprite where we want it
	// regardless of the sprites size, we know that when we draw the sprite at
	// a point on the map, that is where it's anchor point will be.
	x = (float)(g_sprite.pos_x - g_sprite.anchor_x);
	y = (float)(g_sprite.pos_y - g_sprite.anchor_y);
	world_to_screen( &x, &y );

	// create the destination rectangle
	dest.x = (Uint16)x;
	dest.y = (Uint16)y;
	dest.w = SPRITE_WIDTH;
	dest.h = SPRITE_HEIGHT;

	// SDL_BlitSurface could be exchanged for your favorite drawing API
	SDL_BlitSurface( g_image2, &g_sprite.rect[g_sprite.dir][g_sprite.frame],
		g_screen, &dest );
}

// the name says it all here as well
// the camera here is set up for the wrapping effect
// you would add boundaries for it in here if you wanted to limit the camera movement
void move_camera()
{
	// the camera is always centered on the player
	// you would add mins and maxes in here if you
	// were not wrapping and you wanted the map not to scroll
	// when you reach a certain boundary
	g_camera.x = (float)(g_sprite.pos_x - SCREEN_WIDTH/2);

	if ( g_camera.x < 0 )
		g_camera.x = g_camera.x + MAP_WIDTH*TILE_WIDTH;

	g_camera.y = (float)(g_sprite.pos_y - SCREEN_HEIGHT/2);

	if ( g_camera.y < 0 )
		g_camera.y = g_camera.y + MAP_HEIGHT*TILE_HEIGHT;
}

// basic handling of SDL events
// returns true unless the escape key has been
// pressed or the window has been closed in which case
// it returns false.
bool handle_input()
{
	SDL_Event event;

	while( SDL_PollEvent( &event ) )
	{
		switch( event.type )
		{
		// Just as a way to show world to screen and screen to world
		// coordinates.
		// uncomment it to see where you click on the map
		/*
		case SDL_MOUSEBUTTONDOWN:
			switch( event.button.button )
			{
			case SDL_BUTTON_LEFT:
			{
				float worldx = event.button.x, worldy = event.button.y;

				screen_to_world( &worldx, &worldy );

				printf( "screen (%d,%d)  world (%.0f,%.0f)  sprite (%.2f,%.2f)\n\r",
					event.button.x, event.button.y,
					worldx, worldy,
					g_sprite.pos_x, g_sprite.pos_y );
				break;
			}

			case SDL_BUTTON_MIDDLE:
			case SDL_BUTTON_RIGHT:
				break;
			}
			break;
		*/
		case SDL_KEYDOWN:
			switch( event.key.keysym.sym )
			{
				// escape is our exit key but could be handled differently
				case SDLK_ESCAPE:
					return FALSE;

				// instead of using SDL_GetKeyState (although that is probably faster)
				// we react to the SDL_KEYDOWN events and fill our "keypad" array.
				// pretty straight forward.
				case SDLK_DOWN:
					g_keys[GAME_DOWN] = TRUE;
					break;

				case SDLK_UP:
					g_keys[GAME_UP] = TRUE;
					break;

				case SDLK_RIGHT:
					g_keys[GAME_RIGHT] = TRUE;
					break;

				case SDLK_LEFT:
					g_keys[GAME_LEFT] = TRUE;
					break;

				default: break;
			}
			break;

			case SDL_KEYUP:
			switch( event.key.keysym.sym )
			{
				// instead of using SDL_GetKeyState (although that is probably faster)
				// we react to the SDL_KEYUP events and fill our "keypad" array.
				// pretty straight forward.
				case SDLK_DOWN:
					g_keys[GAME_DOWN] = FALSE;
					break;

				case SDLK_UP:
					g_keys[GAME_UP] = FALSE;
					break;

				case SDLK_RIGHT:
					g_keys[GAME_RIGHT] = FALSE;
					break;

				case SDLK_LEFT:
					g_keys[GAME_LEFT] = FALSE;
					break;

				default: break;
			}
			break;

		// our window had been closed
		case SDL_QUIT:
			return FALSE;

		default:
			break;
		}
	}

	return TRUE;
}

// ummm.... 
void draw_map()
{
	Sint32 x, y, off_x, off_y, tile_x, tile_y;
	TILE *tile;
	SDL_Rect rect;

	// this is the offset into the tile we are on
	off_x = -(Sint32)g_camera.x % TILE_WIDTH;
	off_y = -(Sint32)g_camera.y % TILE_HEIGHT;

	// this cuold be optimized a bunch but I left the division in there
	// to show what is going on.
	// the overdraw is there to make sure we don't get seams on the edges
	for ( y = 0; y <= SCREEN_HEIGHT/TILE_HEIGHT + /*overdraw*/2; y++ )
	{
		for ( x = 0; x <= SCREEN_WIDTH/TILE_WIDTH + /*overdraw*/2; x++ )
		{
			// figure out which tile we want to draw based
			// on the camera position.
			// By doing this we could detatch the camera from the player
			// and move the camera independently and the rendering doesn't
			// care about the player. Think like warcraft moving the camera.
			tile_x = (Sint32)(g_camera.x/TILE_WIDTH + x);
			tile_y = (Sint32)(g_camera.y/TILE_HEIGHT + y);

			tile = get_tile( tile_x, tile_y );

			// here is where we figure out where to draw the tile
			rect.x = off_x + x*TILE_WIDTH;
			rect.y = off_y + y*TILE_HEIGHT;
			rect.w = TILE_WIDTH;
			rect.h = TILE_HEIGHT;

			// draw
			SDL_BlitSurface( g_image, &tile->rect, g_screen, &rect );
		}
	}
}

void game_init()
{
	// init SDL
	SDL_Init( SDL_INIT_VIDEO );

	// seed the random number generator
	srand( SDL_GetTicks() );

	// init the screen format
	// saving a pointer to the screen isn't necessary with SDL
	// SDL_GetVideoSurface() may be more appropriate in some cases
	// where you resize the screen or switch between full/windowed mode
	// and the screen surface is re-evaluated.
	g_screen = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, SCREEN_FLAGS );

	// initiate our tile surface
	load_tiles( "tiles.bmp" );

	// generate our map
	load_tilemap( MAP_WIDTH, MAP_HEIGHT );

	// load our player sprite
	load_sprite( "sprite.bmp" );

	// the camera is at 0,0
	g_camera.x = 0;
	g_camera.y = 0;

	// no keys are pressed
	for ( int key = 0; key < 4; key++ )
		g_keys[key] = FALSE;
}

// the main game loop
void game_loop()
{
	int now_time, last_time;
	Uint32 total_frames = 0; 

	// initiate the timer
	now_time = last_time = SDL_GetTicks();

	// run inside this loop while we haven't closed the window
	// or pressed the escape key
	while( handle_input() )
	{
		// set the now time
		now_time = SDL_GetTicks();

		// clear the screen
		SDL_FillRect( g_screen, NULL, 0 );

		// update the sprite base don how much time has passed
		update_sprite( (float)(now_time - last_time) );
		// move the camera to focus on the sprite
		move_camera();

		// draw the map
		// with animated tiles you would place the 
		draw_map();

		// draw the sprite
		draw_sprite();

		// Update the display
		SDL_Flip( g_screen );

		// set the last time
		last_time = now_time;

		// frames
		total_frames++;
	}

	// report the frames
	printf( "total time: %-4.2f sec  total frames: %ld   avg fps: %ld\n\r",
		(float)(last_time/1000),
		total_frames,
		total_frames/(last_time/1000) );
}

// make sure all of our game resources are released
void game_cleanup()
{
	SDL_FreeSurface( g_image );
	g_image = NULL;

	delete [] g_tilemap;
	g_tilemap = NULL;

	SDL_Quit();
}


// funny how the main portion of the game can be so small.
int main( int args, char *argv[] )
{
	game_init();

	game_loop();

	game_cleanup();

	return 0;
}